using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;
using Infragistics.Windows.DataPresenter;
using Infragistics.Windows.DataPresenter.Events;

namespace Infragistics.Extensions
{
    public class FieldAdornerSettings
    {
        #region Members

        private DataPresenterBase _presenter;
        private UIElementAdorner<Control> _adorner;

        #endregion //Members

        #region Constructors

        public FieldAdornerSettings()
        {
            FieldAdorners = new FieldAdornerCollection();
        }

        #endregion //Constructors

        #region Properties

        protected bool AdornerIsVisible
        {
            get { return _adorner.Visibility == Visibility.Visible; }
        }

        public FieldAdornerCollection FieldAdorners { get; set; }

        #endregion //Properties

        public void Attach(DataPresenterBase presenter)
        {
            _presenter = presenter;
            _adorner = new UIElementAdorner<Control>(_presenter);
            HookEventHandlers();
        }

        public void Detach()
        {
            UnhookEventHandlers();
        }

        private void HookEventHandlers()
        {
            _presenter.CellChanged += CellChanged;
            _presenter.CellActivated += CellActivated;
            _presenter.EditModeEnded += EditModeEnded;
            _presenter.EditModeStarted += EditModeStarted;
            _presenter.GroupByArea.Collapsed += GroupByAreaExpansionStateChanged;
            _presenter.GroupByArea.Expanded += GroupByAreaExpansionStateChanged;
            _presenter.Grouped += Grouped;
            _presenter.IsKeyboardFocusWithinChanged += IsKeyboardFocusWithinChanged;
            _presenter.PreviewKeyDown += PreviewKeyDown;
            _presenter.RecordActivated += RecordActivated;
            _presenter.RecordAdded += RecordAdded;
            _presenter.RecordCollapsed += RecordCollapsed;
            _presenter.RecordExpanded += RecordExpanded;
            _presenter.RecordsInViewChanged += RecordsInViewChanged;
            _presenter.SizeChanged += SizeChanged;
            _presenter.Sorted += Sorted;
        }

        private void UnhookEventHandlers()
        {
            _presenter.CellChanged -= CellChanged;
            _presenter.CellActivated -= CellActivated;
            _presenter.EditModeEnded -= EditModeEnded;
            _presenter.EditModeStarted -= EditModeStarted;
            _presenter.GroupByArea.Collapsed -= GroupByAreaExpansionStateChanged;
            _presenter.GroupByArea.Expanded -= GroupByAreaExpansionStateChanged;
            _presenter.Grouped -= Grouped;
            _presenter.IsKeyboardFocusWithinChanged -= IsKeyboardFocusWithinChanged;
            _presenter.PreviewKeyDown -= PreviewKeyDown;
            _presenter.RecordActivated -= RecordActivated;
            _presenter.RecordAdded -= RecordAdded;
            _presenter.RecordCollapsed -= RecordCollapsed;
            _presenter.RecordExpanded -= RecordExpanded;
            _presenter.RecordsInViewChanged -= RecordsInViewChanged;
            _presenter.SizeChanged -= SizeChanged;
            _presenter.Sorted -= Sorted;
        }

        #region Event Handlers

        // This method is invoked when the value of a cell is modified.
        void CellChanged(object sender, CellChangedEventArgs e)
        {
            VerifyAdorner();
        }

        void CellActivated(object sender, CellActivatedEventArgs e)
        {
            VerifyAdorner();
        }

        void EditModeEnded(object sender, EditModeEndedEventArgs e)
        {
            VerifyAdorner();
        }

        void EditModeStarted(object sender, EditModeStartedEventArgs e)
        {
            VerifyAdorner();
        }

        void GroupByAreaExpansionStateChanged(object sender, RoutedEventArgs e)
        {
            MoveAdornerWhileGroupByAreaExpansionStateChanges();
        }

        void Grouped(object sender, GroupedEventArgs e)
        {
            VerifyAdornerAsync();
        }

        void IsKeyboardFocusWithinChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            // This is necessary for when the grid loses focus, and then 
            // gets focus back again.  We need to see if the adorning
            // editor should be displayed for the active cell.
            bool shouldAttemptToShowEditor =
                _presenter.IsKeyboardFocusWithin &&
                _presenter.ActiveCell != null &&
                Keyboard.FocusedElement == _presenter;

            if (shouldAttemptToShowEditor)
                this.VerifyAdorner();
        }

        void PreviewKeyDown(object sender, KeyEventArgs e)
        {
            e.Handled = ProcessKeyStroke(e.Key, e.KeyboardDevice);
        }

        void RecordActivated(object sender, RecordActivatedEventArgs e)
        {
            VerifyAdorner();
        }

        void RecordAdded(object sender, RecordAddedEventArgs e)
        {
            VerifyAdornerAsync();
        }

        void RecordCollapsed(object sender, RecordCollapsedEventArgs e)
        {
            VerifyAdornerAsync();
        }

        void RecordExpanded(object sender, RecordExpandedEventArgs e)
        {
            VerifyAdornerAsync();
        }

        void RecordsInViewChanged(object sender, RecordsInViewChangedEventArgs e)
        {
            VerifyAdorner();
        }

        void SizeChanged(object sender, SizeChangedEventArgs e)
        {
            RelocateAdorner();
        }

        void Sorted(object sender, SortedEventArgs e)
        {
            _presenter.Dispatcher.BeginInvoke(DispatcherPriority.Background, (Action)(() =>
            {
                if (_adorner.Child != null)
                    RelocateAdorner();
                else
                    VerifyAdorner();
            }));
        }

        #endregion // Event Handlers

        #region Private Helpers

        #region GiveFocusToEditor

        void GiveFocusToEditor()
        {
            if (_adorner.Child == null || !_adorner.Child.IsTabStop)
                return;

            // Setting focus to the adorner first ensures that the controls
            // inside the editor can receive focus properly.
            FocusManager.SetFocusedElement(_presenter, _adorner);
            FocusManager.SetFocusedElement(_presenter, _adorner.Child);
        }

        #endregion // GiveFocusToEditor

        #region HideAdorner

        void HideAdorner()
        {
            if (AdornerIsVisible)
            {
                _adorner.Visibility = Visibility.Collapsed;
                _adorner.Child = null;
            }
        }

        #endregion // HideAdorner

        #region MoveAdornerToActiveCell

        bool MoveAdornerToActiveCell(Control adorningEditor)
        {
            if (_presenter.ActiveCell == null)
                return false;

            if (adorningEditor != null && _adorner.Child != adorningEditor)
                _adorner.Child = adorningEditor;

            CellValuePresenter cvp = GetCellValuePresenterForActiveCell();
            if (cvp != null)
            {
                Point location = CalculateAdornerLocation(cvp);
                _adorner.SetOffsets(location.X, location.Y);
                return true;
            }

            return false;
        }

        CellValuePresenter GetCellValuePresenterForActiveCell()
        {
            Cell activeCell = _presenter.ActiveCell;

            if (activeCell == null)
                return null;

            CellValuePresenter cvp = CellValuePresenter.FromCell(activeCell);

            if (cvp == null || !cvp.IsDescendantOf(_presenter) || !cvp.IsVisible)
                return null;

            return cvp;
        }

        Point CalculateAdornerLocation(CellValuePresenter cvp)
        {
            GeneralTransform trans = cvp.TransformToAncestor(_presenter);
            Point location = trans.Transform(new Point(0, 0));

            VerifyAdornerMetrics();

            IScrollInfo scrollInfo = _presenter.ScrollInfo;

            // Make sure the adorner does not extend past the grid's bottom edge.
            bool isScrollingHorizontally = scrollInfo.ViewportWidth < scrollInfo.ExtentWidth;
            double horizontalScrollbarHeight = isScrollingHorizontally ? SystemParameters.ScrollHeight : 0;

            double bottomEdgeOfAdorner = location.Y + cvp.ActualHeight + _adorner.ActualHeight;
            if ((_presenter.ActualHeight - horizontalScrollbarHeight) < bottomEdgeOfAdorner)
                location.Offset(0, -_adorner.ActualHeight);
            else
                location.Offset(0, cvp.ActualHeight);

            // Make sure the adorner does not extend past the grid's right edge.
            bool isScrollingVertically = scrollInfo.ViewportHeight < scrollInfo.ExtentHeight;
            double verticalScrollbarWidth = isScrollingVertically ? SystemParameters.ScrollWidth : 0;

            double rightEdgeOfAdorner = location.X + _adorner.ActualWidth;
            double horizontalDiff = rightEdgeOfAdorner - (_presenter.ActualWidth - verticalScrollbarWidth);
            if (0 < horizontalDiff)
                location.Offset(-horizontalDiff, 0);

            return location;
        }

        void VerifyAdornerMetrics()
        {
            VerifyAdornerLayer();

            if (!AdornerIsVisible)
                _adorner.Visibility = Visibility.Visible;

            // Force the adorner through a measure pass since the hosted editor control
            // might have changed since the last time the metrics were calculated.
            _adorner.InvalidateMeasure();
            _adorner.UpdateLayout();
        }

        #endregion // MoveAdornerToActiveCell

        #region MoveAdornerWhileGroupByAreaExpansionStateChanges

        void MoveAdornerWhileGroupByAreaExpansionStateChanges()
        {
            if (!AdornerIsVisible)
                return;

            DispatcherTimer timer = new DispatcherTimer();
            timer.Interval = TimeSpan.FromMilliseconds(1);
            int counter = 0;
            timer.Tick += delegate
            {
                if (++counter > 500)
                {
                    timer.Stop();
                    timer = null;
                }
                else if (AdornerIsVisible)
                {
                    MoveAdornerToActiveCell(null);
                }
            };
            timer.Start();
        }

        #endregion // MoveAdornerWhileGroupByAreaExpansionStateChanges

        #region ProcessKeyStroke

        bool ProcessKeyStroke(Key key, KeyboardDevice device)
        {
            if (!AdornerIsVisible)
                return false;

            bool giveEditorFocus =
                key == Key.Tab &&
                device.Modifiers != ModifierKeys.Shift &&
                _presenter.ActiveCell != null &&
                _adorner.Child != null &&
                _adorner.Child.IsTabStop;

            if (!giveEditorFocus)
                return false;

            GiveFocusToEditor();

            return true;
        }

        #endregion // ProcessKeyStroke

        #region RelocateAdorner

        void RelocateAdorner()
        {
            if (AdornerIsVisible)
            {
                bool resumeEditing = false;

                if (_adorner.IsKeyboardFocusWithin)
                {
                    resumeEditing = true;

                    // We must take the adorner out of focus otherwise it won't move.
                    _presenter.Focus();
                }

                // Delay the call that moves the adorner so that the
                // grid has a chance to finish updating its elements.
                _presenter.Dispatcher.BeginInvoke(
                    DispatcherPriority.Background,
                    (Action)delegate
                    {
                        MoveAdornerToActiveCell(null);

                        if (resumeEditing)
                            GiveFocusToEditor();
                    });
            }
        }

        #endregion // RelocateAdorner

        void VerifyAdorner()
        {
            //if the cell is a filter cell don't do anything with the adorner
            if (_presenter.ActiveCell is FilterCell)
                return;

            VerifyAdornerLayer();

            FrameworkElement adorningEditor = null;
            if (_presenter.ActiveCell != null)
            {
                //TODO: extract method
                var fieldAdorner = FieldAdorners[_presenter.ActiveCell.Field.Name];
                if (fieldAdorner != null)
                {
                    adorningEditor = fieldAdorner.Adorner;

                    var iFieldAdorner = adorningEditor as IXamDataPresenterFieldAdorner;
                    if (iFieldAdorner != null && !iFieldAdorner.IsAttached)
                        iFieldAdorner.Attach(_presenter);

                    adorningEditor.DataContext = _presenter.ActiveCell.Record.DataItem;
                }
            }

            bool movedAdorner = false;
            if (adorningEditor != null)
                movedAdorner = MoveAdornerToActiveCell((Control)adorningEditor);

            if (!movedAdorner)
                HideAdorner();
        }

        void VerifyAdornerAsync()
        {
            _presenter.Dispatcher.BeginInvoke(DispatcherPriority.Background, (Action)(() => VerifyAdorner()));
        }

        /// <summary>
        /// Ensures that the adorner is in the adorner layer.
        /// </summary>
        /// <returns>True if the adorner is in the adorner layer, else false.</returns>
        bool VerifyAdornerLayer()
        {
            if (_adorner.Parent != null)
                return true;

            AdornerLayer layer = AdornerLayer.GetAdornerLayer(_presenter);
            if (layer == null)
                return false;

            layer.Add(_adorner);
            return true;
        }

        #endregion // Private Helpers
    }
}
