﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Infragistics.Windows.DataPresenter;

namespace Infragistics.Extensions
{
    public class XamDataPresenterFieldAdorner : ContentControl, IXamDataPresenterFieldAdorner
    {
        #region Members

        private DataPresenterBase _presenter;
        private Cell _adornedCell;

        #endregion // Members

        #region Constructors

        static XamDataPresenterFieldAdorner()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(XamDataPresenterFieldAdorner), new FrameworkPropertyMetadata(typeof(XamDataPresenterFieldAdorner)));
        }

        public XamDataPresenterFieldAdorner()
        {
            // Hide the control when the user clicks a Close button.
            this.CommandBindings.Add(new CommandBinding(ApplicationCommands.Close, Close, CanClose));

            // When the DataContext changes, that means this control
            // is being shown for a different row than before.  In that
            // case we have to clean out some previous settings associated
            // with the previously adorned cell.
            DataContextChanged += new DependencyPropertyChangedEventHandler(XamDataPresenterFieldAdorner_DataContextChanged);
        }

        #endregion // Constructors

        #region Properties

        public static readonly DependencyProperty CaptionProperty = DependencyProperty.Register("Caption", typeof(string), typeof(XamDataPresenterFieldAdorner), new UIPropertyMetadata(null));
        public string Caption
        {
            get { return (string)GetValue(CaptionProperty); }
            set { SetValue(CaptionProperty, value); }
        }

        public bool IsClosed
        {
            get { return base.Visibility != Visibility.Visible; }
            private set
            {
                base.Visibility = value ? Visibility.Collapsed : Visibility.Visible;

                if (value && base.IsKeyboardFocusWithin)
                    _presenter.Focus();
            }
        }

        #endregion //Properties

        #region Base Class Overrides

        protected override void OnPreviewKeyDown(KeyEventArgs e)
        {
            RoutedCommand cmd = this.GetCommandForKey(e.Key, e.KeyboardDevice);

            if (cmd != null)
            {
                e.Handled = true;

                _presenter.ExecuteCommand(cmd);

                // Set focus to the grid because our TextBox might have focus now.
                // This is necessary because none of the grid cells are editable.
                _presenter.Focus();
            }

            base.OnPreviewKeyDown(e);
        }

        protected override void OnPreviewMouseDown(MouseButtonEventArgs e)
        {
            // If the user clicks on a TextBox, give it focus and select all the text it contains.
            DependencyObject depObj = e.OriginalSource as DependencyObject;
            TextBox txt = Infragistics.Windows.Utilities.GetAncestorFromType(depObj, typeof(TextBox), false, this) as TextBox;
            if (txt != null && !txt.IsFocused)
            {
                txt.Focus();
                txt.SelectAll();
                e.Handled = true;
            }

            base.OnPreviewMouseDown(e);
        }

        #endregion //Base Class Overrides

        #region Event Handlers

        private void XamDataPresenterFieldAdorner_DataContextChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            if (IsClosed)
                IsClosed = false;

            _adornedCell = _presenter.ActiveCell;
        }

        #endregion //Event Handlers

        #region Commands

        private void Close(object sender, ExecutedRoutedEventArgs e)
        {
            IsClosed = true;
        }

        private void CanClose(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = e.Handled = true;
        }

        #endregion //Commands

        #region Methods

        private RoutedCommand GetCommandForKey(Key key, KeyboardDevice device)
        {
            RoutedCommand cmd = null;

            if (key == Key.Tab)
            {
                cmd = device.Modifiers == ModifierKeys.Shift ?
                    // If the Shift key is pressed, give focus back to the active cell.
                    DataPresenterCommands.StartEditMode :
                    // If the user only pressed Tab, move to the next cell.
                    DataPresenterCommands.CellNextByTab;
            }
            else if (!base.IsTabStop)
            {
                // If this control is not a tab stop, it is still possible
                // that the user clicked on a child control to give it focus.
                // In that case, we must manually move focus to the correct cell.
                if (key == Key.Right)
                {
                    cmd = DataPresenterCommands.CellRight;
                }
                else if (key == Key.Left)
                {
                    cmd = DataPresenterCommands.CellLeft;
                }
                else if (key == Key.Up)
                {
                    cmd = DataPresenterCommands.CellAbove;
                }
                else if (key == Key.Down)
                {
                    cmd = DataPresenterCommands.CellBelow;
                }
                else if (key == Key.Escape)
                {
                    // Move input focus back to the active cell.
                    cmd = DataPresenterCommands.StartEditMode;
                }
            }

            return cmd;
        }

        #endregion //Methods

        #region IXamDataPresenterFieldAdorner

        public bool IsAttached { get; set; }

        public void Attach(DataPresenterBase presenter)
        {
            if (presenter == null)
                throw new ArgumentNullException("presenter");

            _presenter = presenter;
            _presenter.CellActivated += (sender, e) =>
            {
                if (this.IsClosed && e.Cell != _adornedCell)
                    this.IsClosed = false;
            };

            IsAttached = true;
        }

        #endregion //IXamDataPresenterFieldAdorner
    }
}
